@@ -1,5 +1,5 @@
 import random
-from typing import List, Tuple
+from typing import List, Tuple, Iterable
 
 from smallgraphlib import (
     Graph,
@@ -16,6 +16,7 @@ from smallgraphlib.core import (
     _TIKZ_EXPORT_MAX_MULTIPLE_LOOPS_SUPPORT,
     AbstractGraph,
     Edge,
+    Node,
 )
 from smallgraphlib.utilities import Multiset
 
@@ -79,7 +80,6 @@ def perfect_binary_tree(height: int):
     """
     nodes: List[int] = []
     edges: List[Edge] = []
-    i = 1
     for level in range(height):
         nodes.extend(range(1 << level, 1 << level + 1))
     for level in range(height - 1):
@@ -100,6 +100,7 @@ def random_graph(
     max_multiple_edges: int = None,
     max_multiple_loops: int = None,
     shuffle_nodes=False,
+    nodes_names: Iterable[Node] = None,
 ) -> AbstractGraph:
     """Create a random graph satisfying given constraints.
 
@@ -119,7 +120,6 @@ def random_graph(
                 f"Degree is at most {max_degree_for_simple_graph} "
                 f"for a simple {graph_type} graph of order {order}."
             )
-    if simple:
         if max_multiple_edges is not None or max_multiple_loops is not None:
             raise ValueError(
                 "Conflicting arguments: `simple` must be set to False, "
@@ -127,24 +127,33 @@ def random_graph(
             )
         max_multiple_edges = 1
         max_multiple_loops = 0
+
     elif tikz_export_supported:
-        if max_multiple_edges is not None or max_multiple_loops is not None:
-            raise ValueError(
-                "Conflicting arguments: `tikz_export_supported` must be set to False, "
-                "else `max_multiple_edges` and `max_multiple_loops` will be ignored."
+        if max_multiple_edges is None:
+            max_multiple_edges = _TIKZ_EXPORT_MAX_MULTIPLE_EDGES_SUPPORT
+        elif max_multiple_edges > _TIKZ_EXPORT_MAX_MULTIPLE_EDGES_SUPPORT:
+            print(
+                "Warning: if `tikz_export_supported` is True, `max_multiple_edges` can't be set "
+                f"to more than {_TIKZ_EXPORT_MAX_MULTIPLE_EDGES_SUPPORT}."
             )
-        max_multiple_edges = _TIKZ_EXPORT_MAX_MULTIPLE_EDGES_SUPPORT
-        max_multiple_loops = _TIKZ_EXPORT_MAX_MULTIPLE_LOOPS_SUPPORT
+            max_multiple_edges = _TIKZ_EXPORT_MAX_MULTIPLE_EDGES_SUPPORT
+
+        if max_multiple_loops is None:
+            max_multiple_loops = _TIKZ_EXPORT_MAX_MULTIPLE_LOOPS_SUPPORT
+        elif max_multiple_loops > _TIKZ_EXPORT_MAX_MULTIPLE_LOOPS_SUPPORT:
+            print(
+                "Warning: if `tikz_export_supported` is True, `max_multiple_loops` can't be set "
+                f"to more than {_TIKZ_EXPORT_MAX_MULTIPLE_LOOPS_SUPPORT}."
+            )
+            max_multiple_loops = _TIKZ_EXPORT_MAX_MULTIPLE_LOOPS_SUPPORT
+
     else:
         if max_multiple_edges is None:
             max_multiple_edges = degree
         if max_multiple_loops is None:
             max_multiple_loops = degree
 
-    if directed:
-        max_degree = order * (max_multiple_loops + (order - 1) * max_multiple_edges)
-    else:
-        max_degree = order * max_multiple_loops + (order * (order - 1) * max_multiple_edges) // 2
+    max_degree = order * max_multiple_loops + (order * (order - 1) * max_multiple_edges) // 2
 
     if degree > max_degree:
         raise ValueError(f"Degree must not exceed {max_degree} with given contraints.")
@@ -152,23 +161,21 @@ def random_graph(
     nodes: List[int] = list(range(1, order + 1))
     edges: List[Tuple[int, int]] = []
 
+    # fmt: off
     # Keep track of remaining edges possibilities, for each (start, end) nodes couple.
+    # Only nodes which accept edges will be kept (all except the last one
+    # for undirected simple graph, since start < end).
     counters = {
         start: counter
         for start in nodes
-        if (
-            counter := Multiset(
-                {
-                    end: (max_multiple_loops if start == end else max_multiple_edges)
-                    for end in nodes
-                    if directed or start <= end  # for undirected graph, only keep start <= end
-                }
-            )
-        ).total()
-        > 0
+        if (counter := Multiset(
+            {
+                end: (max_multiple_loops if start == end else max_multiple_edges)
+                for end in nodes
+                if directed or start <= end  # for undirected graph, only keep start <= end
+            })).total() > 0
     }
-    # Keep only nodes which accept edges (all except the last one
-    # for undirected simple graph, since start < end).
+    # fmt: on
     starts = list(counters)
 
     while len(edges) < degree:
@@ -181,7 +188,20 @@ def random_graph(
             # Don't select anymore this node as start point: we can't add any other edge to it.
             starts.remove(start)
         edges.append((start, end))  # Loop will always end. :)
+        if tikz_export_supported and directed and start != end:
+            # Edges of opposite directions are not strictly speaking multiple edges,
+            # but as far as Tikz is concerned, they must count as multiple edges since
+            # they are drawn parallely (it's better to avoid <-> arrows, i.e. arrows with two heads,
+            # since it's too easy for students to miss an arrow tail on a node then).
+            # So we must decrease counter for opposite edge too.
+            counter = counters[end]
+            assert counter[start] > 0
+            counter[start] -= 1
+            if counter.total() == 0:
+                starts.remove(end)
     g = graph(nodes, *edges, directed=directed)
     if shuffle_nodes:
         g.shuffle_nodes()
+    if nodes_names:
+        g.rename_nodes(dict(enumerate(list(nodes_names)[: len(g.nodes)], start=1)))
     return g
